
from . import _share


__all__ = ()


def unit(info, root, data, flush = False):

    left = set(root if flush else ())

    for (name, subdata) in data.items():

        try:
            field = info[name]
        except KeyError:
            continue

        left.discard(name)

        try:
            value = root[name]
        except KeyError:
            value = root[name] = field.type(subdata)
            continue

        try:
            any_(value, subdata)
        except ValueError:
            root[name] = subdata

    for name in left:
        try:
            del data[name]
        except KeyError:
            continue


def list_(compare, build, root, data, flush = True):

    rindex = 0

    while True:
        try:
            value = root[rindex]
        except IndexError:
            break
        for (dindex, subdata) in enumerate(data):
            if not compare(value, subdata):
                continue
            break
        else:
            if flush:
                del root[index]
            continue
        any_(value, subdata)
        del data[dindex]
        rindex += 1

    root.extend(map(build, data))


def dict_(identify, build, root, data, flush = False):

    if identify is None:
        identities = list(data)
        data = data.values()
    else:
        identities = list(map(identify, data))

    for (key, value) in tuple(root.items()):
        for (index, subdata) in enumerate(data):
            identity = identities[index]
            if not key == identity:
                continue
            break
        else:
            if flush:
                del root[key]
            continue
        any_(value, subdata)
        del data[index], identities[index]

    root.update(zip(identities, map(build, data)))


def any_(value, data, **kwargs):

    context = _share.get(value.__class__)

    root = context.datas[value]

    context.update(root, data, **kwargs)



import collections


__all__ = ()


Context = collections.namedtuple(
    'Context',
    'type update datas'
)


contexts = []


def get(type):

    for context in contexts:
        if not issubclass(type, context.type):
            continue
        break
    else:
        raise ValueError('not found')

    return context




from . import _builds
from . import _updates


__all__ = ('Field', 'missing', 'ObjectBase', 'object', 'ListBase', 'list',
           'DictBase', 'dict', 'update')


class Field(_builds.Field):

    """
    Field(type=None, make=None, name=None)

    Describes how an :func:`.object` instance key should be handled.

    :param func type:
        Called with ``(data)`` for creation of new value.
    :param func make:
        Called with ``self`` upon attribute access. Close to :class:`property`.
    :param str name:
        Exclusive alternative to accessing an data value.
    """

    __slots__ = ()


missing = _builds.missing
"""
Returned upon accessing valid but empty attribute. Exhibits negative booleanity.
"""


class ObjectBase:

    """
    Default ABC for :func:`.object` classes.
    """

    __slots__ = ()


def object(*args, cls = ObjectBase, **kwargs):

    """
    object(info, identify=None, cls=ObjectBase, **behave)

    Create a class supporting dynamically updated data.

    :param dict[str,Field] info:
        Determines how to handle incoming data.
    :param func identify:
        Should return a hash for caching.
    :param type cls:
        Base class for created class.

    .. code-block:: py

        User_info = {
            'id': vessel.Field(
                type = int
            ),
            'username': vessel.Field(
                type = str,
                name = 'name'
            ),
            'tag': vessel.Field(
                type = str
            ),
            'display': vessel.Field(
                make = lambda self: f'{self.name}#{self.tag}'
            )
        }

        User_identify = lambda data: int(data['id'])

        User = vessel.object(User_info, User_identify)

        user = User(
            {
                'id': 0,
                'username': 'Exa',
                'tag': '0000'
            }
        )

        print(user.display)

    Above snippet prints ``Exa#0000``.

    .. note::

        When calling generated classes, the **same** instance will be returned
        for datas whose identity matches an existing one. To disable this
        behavior temporarily, use ``unique = True``. This is incosequential for
        classes generated without ``identify``.

        .. code-block:: py

            data = {'id': 0, ...}
            user0 = User(data)
            user1 = User(data)
            user2 = User(data, unique = True)
            print(user0 is user1)
            print(user0 is user2)

        Above snipper prints ``True`` and ``False``
    """

    return _builds.build_vessel(*args, cls = cls, **kwargs)


class ListBase:

    """
    Default ABC for :func:`.list` classes.
    """

    __slots__ = ()


def list(*args, cls = ListBase, **kwargs):

    """
    list(type, compare, cls=ListBase, **behave)

    Create a class supporting :class:`list` access operations hosting objects.

    :param func type:
        Called with ``(data)`` for creation of new value.
    :param func compare:
        Called with ``(root, data)`` and should return :class:`bool` for
        whether their identities match.
    :param type cls:
        Base class for created class.

    .. code-block::

        User_compare = lambda self, data: self.id == User_identify(data)

        Account_info = {
            'owner': vessel.Field(
                type = User
            ),
            'friends': vessel.Field(
                type = vessel.list(User, User_compare)
            )
        }

        Account = vessel.object(Account_info)

        account = Account(
            {
                'owner': {
                    'id': 0,
                    'username': 'Exa',
                    'tag': '0000',
                },
                'friends': [
                    {
                        'id': 1,
                        'username': 'Rob',
                        'tag': '1234'
                    },
                    {
                        'id': 2,
                        'username': 'Nil',
                        'tag': '5678'
                    }
                ]
            }
        )

        for friend in account.friends:
            print(friend.display)

    Above snippet will print ``Rob#1234`` and ``Nil#5678``.
    """

    return _builds.build_list(*args, cls = ListBase, **kwargs)


class DictBase:

    """
    Default ABC for :func:`.dict` classes.
    """

    __slots__ = ()


def dict(*args, cls = DictBase, **kwargs):

    """
    dict(type, identify, cls=DictBase, **behave)

    Create a class supporting :class:`dict` access operations hosting objects.

    :param func type:
        Called with ``(data)`` for creation of new value.
    :param func compare:
        Called with ``(data)`` and should return respective key.
    :param type cls:
        Base class for created class.

    .. code-block:: py

        Account_info = {
            'owner': vessel.Field(
                type = User
            ),
            'friends': vessel.Field(
                type = vessel.dict(User, User_identify, flush = True)
            )
        }

        account = Account({...})

        for friend in account.friends:
            print(friend.display)

        print(account.friends[1].name)

    Above snippet will print ``Rob#1234``, ``Nil#5678`` and ``Rob``.

    .. note::

        Unlike normal :class:`dict`\s, direct iteration yields values instead of
        keys.
    """

    return _builds.build_dict(*args, cls = DictBase, **kwargs)


def update(*args, **kwargs):

    """
    update(root, data, flush=?)

    Updates an :func:`.object`, :func:`.list` or :func:`.dict` result with new
    data.

    - Mutates data when identity matches with new data.
    - Removes old data when identity does not match with new data if ``flush``.
    - Appends new data when identity does not match with old data.

    :param object root:
        Initial value.
    :param dict or list data:
        Stuff to update root with.
    :param bool flush:
        Whether to remove redundant entries.

    ``flush`` is ``False`` by default except for :func:`.list` results.

    ``data`` **must** be :class:`dict` for :func:`.object` results, and
    :class:`list` otherwise.

    .. note::

        ``**behave`` in other functions overwrites defaults for optional
        arguments.

    .. code-block:: py

        vessel.update(
            account,
            {
                'owner': {
                    'tag': '9012'
                },
                'friends': [
                    {
                        'id': 1,
                        'username': 'Robbert'
                    },
                    {
                        'id': 3,
                        'username': 'Cul',
                        'tag': '3456'
                    }
                ]
            }
        )

        for friend in account.friends:
            print(friend.display)

    Above snippet will print ``Robbert#1234`` and ``Cull#3456``.

    ``Nil`` was removed because ``flush=True`` when using :func:`.dict` for
    ``'friends'``.

    .. note::

      Generated class instances are not designed for mutation outside of
      :func:`.update`.
    """

    return _updates.any_(*args, **kwargs)



import collections
import weakref
import itertools

from . import _share
from . import _updates


__all__ = ('Field', 'missing', 'update', 'build_vessel', 'build_list',
           'build_dict')


def _build(create, update, identify, cls):

    datas = weakref.WeakKeyDictionary()

    if identify:
        units = weakref.WeakValueDictionary()
    else:
        units = None

    class Unit(cls):

        __slots__ = ('__weakref__',)

        def __new__(cls, data, *args, unique = False, **kwargs):

            if identify and not unique:
                identity = identify(data)
                try:
                    self = units[identity]
                except KeyError:
                    forge = attach = True
                else:
                    forge = attach = False
            else:
                forge = True
                attach = False

            if forge:
                self = super().__new__(cls)
                if create:
                    datas[self] = create(data)

            if attach:
                units[identity] = self

            if update:
                _updates.any_(self, data)

            return self

    return Unit, datas


def build(create, update, identify, cls = object):

    Unit, datas = _build(create, update, identify, cls)

    context = _share.Context(Unit, update, datas)
    _share.contexts.append(context)

    return Unit


def _get_data(unit):

    context = _share.get(unit.__class__)

    data = context.datas[unit]

    return data


Field = collections.namedtuple(
    'Field',
    'type name make',
    defaults = (None, None, None)
)


missing = type(
    'missing',
    (),
    {
        '__slots__': (),
        '__bool__': False.__bool__,
        '__repr__': lambda self: ''
    }
)()


def build_vessel(info, identify = None, cls = object, **behave):

    def create(data):

        root = {}

        return root

    def update(root, data, **kwargs):

        kwargs = {**behave, **kwargs}

        _updates.unit(info, root, data, **kwargs)

    names = {field.name: name for (name, field) in info.items() if field.name}

    alias = lambda name: names.get(name, name)

    class Vessel(cls):

        __slots__ = ()

        def __getattr__(self, name):

            name = alias(name)

            try:
                field = info[name]
            except KeyError as error:
                raise AttributeError(*error.args) from None

            context = _share.get(self.__class__)

            if field.make:
                return field.make(self)

            data = context.datas[self]

            value = data.get(name, missing)

            return value

        def __repr__(self):

            data = _get_data(self)

            names = map(alias, data.keys())
            values = data.values()
            items = zip(names, values)

            pairs = ', '.join(map('{0[0]}={0[1]}'.format, items))

            return '{0}({1})'.format(self.__class__.__name__, pairs)

    Unit = build(create, update, identify, Vessel)

    return Unit


def _method_data_proxy_wrap(name):

    def function(self, *args, **kwargs):

        data = _get_data(self)

        return getattr(data, name)(*args, **kwargs)

    return function


_collection_names = ('__iter__', '__len__', '__getitem__')


def _collection__repr__(self):

    data = _get_data(self)

    return '{0}({1})'.format(self.__class__.__name__, len(data))


def _collect(cls, names):

    space = {'__slots__': (), '__repr__': _collection__repr__}

    for name in itertools.chain(_collection_names, names):
        space[name] = _method_data_proxy_wrap(name)

    CollectionBase = type('CollectionBase', (cls,), space)

    return CollectionBase


def build_list(make, compare, cls = object, **behave):

    def create(data):

        root = []

        return root

    def update(root, data, **kwargs):

        kwargs = {**behave, **kwargs}

        _updates.list_(compare, make, root, data, **kwargs)

    CollectionBase = _collect(cls, ())

    Unit = build(create, update, None, CollectionBase)

    class List(Unit):

        __slots__ = ()

    return Unit


def build_dict(make, identify, cls = object, **behave):

    def create(data):

        root = {}

        return root

    def update(root, data, **kwargs):

        kwargs = {**behave, **kwargs}

        _updates.dict_(identify, make, root, data, **kwargs)

    CollectionBase = _collect(cls, ('get', 'keys', 'values', 'items'))

    Unit = build(create, update, None, CollectionBase)

    class Dict(Unit):

        __slots__ = ()

        def __iter__(self):

            yield from self.values()

    return Dict



import setuptools

with open('README.rst') as file:
    readme = file.read()

name = 'vessel'

version = '3.0.0'

author = 'Exahilosys'

url = f'https://github.com/{author}/{name}'

setuptools.setup(
    name = name,
    version = version,
    author = author,
    url = url,
    packages = setuptools.find_packages(),
    license = 'MIT',
    description = 'Dynamic data access.',
    long_description = readme,
    extras_require = {
        'docs': [
            'sphinx'
        ]
    }
)
